"""Module for testing the normalize endpoint."""
import copy
from datetime import datetime

import pytest
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor

from variation.main import normalize as normalize_get_response
from variation.main import to_vrs as to_vrs_get_response
from tests.conftest import assertion_checks


@pytest.fixture(scope="module")
def limk2_gene_context():
    """Create LIMK2 gene context test fixture."""
    return {
        "id": "normalize.gene:LIMK2",
        "type": "GeneDescriptor",
        "label": "LIMK2",
        "gene_id": "hgnc:6614",
        "xrefs": [
            "ncbigene:3985",
            "ensembl:ENSG00000182541"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "name": "approved_name",
                "value": "LIM domain kinase 2",
                "type": "Extension"
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "refseq:NM_016733",
                    "ccds:CCDS33637",
                    "ccds:CCDS13892",
                    "ena.embl:D45906",
                    "uniprot:P53671",
                    "pubmed:10591208",
                    "vega:OTTHUMG00000151251",
                    "omim:601988",
                    "iuphar:2055",
                    "pubmed:8537403",
                    "ccds:CCDS13891",
                    "ucsc:uc003akh.4"
                ]
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "_id": "ga4gh:VCL.IoyhTh4PxvPx8yF9P3IecXDVs_XVbDe9",
                    "type": "ChromosomeLocation",
                    "species_id": "taxonomy:9606",
                    "chr": "22",
                    "interval": {
                        "end": "q12.2",
                        "start": "q12.2",
                        "type": "CytobandInterval"
                    }
                }
            }
        ]
    }


@pytest.fixture(scope="module")
def dis3_p63a():
    """Create DIS3 P63A test fixture."""
    params = {
        "id": "normalize.variation:DIS3%20P63A",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.ueB4LJE1wba68a1w1w6Mrb-MOquy8n3C",
        "variation": {
            "_id": "ga4gh:VA.ueB4LJE1wba68a1w1w6Mrb-MOquy8n3C",
            "location": {
                "_id": "ga4gh:VSL.JcyesAEvndeQYxn4Gzi68hqC1ANZnrGN",
                "interval": {
                    "end": {"value": 63, "type": "Number"},
                    "start": {"value": 62, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.mlWsxfPKINN3o300stAI8oqN5U7P6kEu",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "A",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "gene_context": {
            "id": "normalize.gene:DIS3",
            "type": "GeneDescriptor",
            "label": "DIS3",
            "xrefs": [
                "ensembl:ENSG00000083520",
                "ncbigene:22894"
            ],
            "alternate_labels": [
                "dis3p",
                "RRP44",
                "KIAA1008",
                "2810028N01Rik",
                "EXOSC11"
            ],
            "extensions": [
                {
                    "name": "symbol_status",
                    "value": "approved",
                    "type": "Extension"
                },
                {
                    "name": "approved_name",
                    "value": "DIS3 homolog, exosome endoribonuclease and 3'-5' exoribonuclease",  # noqa: E501
                    "type": "Extension"
                },
                {
                    "name": "chromosome_location",
                    "value": {
                        "species_id": "taxonomy:9606",
                        "interval": {
                            "type": "CytobandInterval",
                            "start": "q21.33",
                            "end": "q21.33"
                        },
                        "_id": "ga4gh:VCL.84IPub_nKl33cWX9pNoPeGsyeVuJnyra",
                        "type": "ChromosomeLocation",
                        "chr": "13"
                    },
                    "type": "Extension"
                },
                {
                    "name": "associated_with",
                    "value": [
                        "vega:OTTHUMG00000017070",
                        "ccds:CCDS9447",
                        "orphanet:470196",
                        "ena.embl:AB023225",
                        "ccds:CCDS45057",
                        "omim:607533",
                        "pubmed:11935316",
                        "refseq:NM_014953",
                        "uniprot:Q9Y2L1",
                        "ccds:CCDS81772",
                        "ucsc:uc001vix.6",
                        "pubmed:9562621"
                    ],
                    "type": "Extension"
                },
                {
                    "name": "previous_symbols",
                    "value": [
                        "KIAA1008"
                    ],
                    "type": "Extension"
                }
            ],
            "gene_id": "hgnc:20604"
        },
        "vrs_ref_allele_seq": "P"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def vhl(vhl_gene_context):
    """Create VHL Tyr185Ter fixture."""
    params = {
        "id": "normalize.variation:NP_000542.1%3Ap.Tyr185Ter",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA._S0nFwX4Y2FPmv5Radf01DAsxQbxA2cc",
        "variation": {
            "_id": "ga4gh:VA._S0nFwX4Y2FPmv5Radf01DAsxQbxA2cc",
            "location": {
                "_id": "ga4gh:VSL._P3rBWI3f7OBs3a4gvZ18QJ6f6dSfqEQ",
                "interval": {
                    "end": {"value": 185, "type": "Number"},
                    "start": {"value": 184, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.z-Oa0pZkJ6GHJHOYM7h5mY_umc0SJzTu",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "*",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001617",
        "vrs_ref_allele_seq": "Y",
        "gene_context": vhl_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def braf_v600e_nucleotide(braf_gene_context, braf_nuc_value):
    """Create a test fixture for BRAF V600E MANE select nucleotide hgvs."""
    variation = copy.deepcopy(braf_nuc_value)
    variation["_id"] = "ga4gh:VA.bsmS0aeGSffgemmymBrpTjTSW2qYeIJE"
    params = {
        "id": "normalize.variation:NM_004333.4%3Ac.1799T%3EA",
        "type": "VariationDescriptor",
        "variation_id": variation["_id"],
        "variation": variation,
        "molecule_context": "transcript",
        "structural_type": "SO:0001483",
        "vrs_ref_allele_seq": "T",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def nm_004448_coding_dna_delins(erbb2_context):
    """Create test fixture for NM_004448.4:c.2326_2327delinsCT."""
    params = {
        "id": "normalize.variation:NM_004448.4%3Ac.2326_2327delinsCT",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.eMxxAEjNduAvg5U3eBZxf0nLtfcMNxqy",
        "variation": {
            "_id": "ga4gh:VA.eMxxAEjNduAvg5U3eBZxf0nLtfcMNxqy",
            "location": {
                "_id": "ga4gh:VSL.bBzTvpLChbWE2SZ7X0drm8NQj5rzNqTK",
                "interval": {
                    "end": {"value": 2502, "type": "Number"},
                    "start": {"value": 2500, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.y9b4LVMiCXpZxOg9Xt1NwRtssA03MwWM",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "CT",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "GG",
        "gene_context": erbb2_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def nc_000007_genomic_delins(braf_gene_context):
    """Create test fixture for NC_000007.13:g.140453135_140453136delinsAT."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.140453135_140453136delinsAT",  # noqa: E501
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.4387UZ6Yssh3XCGKjm71z_WtadpBZT3O",
        "variation": {
            "_id": "ga4gh:VA.4387UZ6Yssh3XCGKjm71z_WtadpBZT3O",
            "location": {
                "_id": "ga4gh:VSL.6PeoFwkO4ISmUjDWoYLkVsATVx8JRApd",
                "interval": {
                    "end": {"value": 2146, "type": "Number"},
                    "start": {"value": 2144, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.I_0feOk5bZ3VfH8ejhWQiMDe9o6o4QdR",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "AT",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "TG",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def nm_000551(vhl_gene_context):
    """Create test fixture for NM_000551.4:c.615delinsAA."""
    params = {
        "id": "normalize.variation:temp",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA._JN_AF5PO9kWKgKxB5T48cypZl7ccEsQ",
        "variation": {
            "_id": "ga4gh:VA._JN_AF5PO9kWKgKxB5T48cypZl7ccEsQ",
            "location": {
                "_id": "ga4gh:VSL.tQRFfWMPPHErWSbKvoEIh2gI2ehHIYAs",
                "interval": {
                    "end": {"value": 685, "type": "Number"},
                    "start": {"value": 684, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.xBKOKptLLDr-k4hTyCetvARn16pDS_rW",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "AA",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "C",
        "gene_context": vhl_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def braf_nuc_value():
    """Create test fixture for BRAF V600E value on c. coordinate."""
    return {
        "location": {
            "_id": "ga4gh:VSL.WnQ4luBGVxa-tMWza5nPzFh1bIeCij3T",
            "interval": {
                "end": {"value": 2025, "type": "Number"},
                "start": {"value": 2024, "type": "Number"},
                "type": "SequenceInterval"
            },
            "sequence_id": "ga4gh:SQ.aKMPEJgmlZXt_F6gRY5cUG3THH2n-GUa",
            "type": "SequenceLocation"
        },
        "state": {
            "sequence": "A",
            "type": "LiteralSequenceExpression"
        },
        "type": "Allele"
    }


@pytest.fixture(scope="module")
def coding_dna_silent_mutation(braf_gene_context, braf_nuc_value):
    """Create test fixture for NM_004333.4:c.1799=."""
    value = copy.deepcopy(braf_nuc_value)
    value["state"]["sequence"] = "T"
    value["_id"] = "ga4gh:VA.kbgr3KgaDXI6jUR8r7A53xlNSN2cgOl8"
    params = {
        "id": "normalize.variation:NM_004333.4%3Ac.1799%3D",
        "type": "VariationDescriptor",
        "variation_id": value["_id"],
        "variation": value,
        "molecule_context": "transcript",
        "structural_type": "SO:0002073",
        "vrs_ref_allele_seq": "T",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def nc_000007_silent_mutation(braf_gene_context, braf_nuc_value):
    """Create test fixture for NC_000007.13:g.140453136=."""
    value = copy.deepcopy(braf_nuc_value)
    value["state"]["sequence"] = "T"
    value["_id"] = "ga4gh:VA.kbgr3KgaDXI6jUR8r7A53xlNSN2cgOl8"
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.140453136%3D",
        "type": "VariationDescriptor",
        "variation_id": value["_id"],
        "variation": value,
        "molecule_context": "transcript",
        "structural_type": "SO:0002073",
        "vrs_ref_allele_seq": "T",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def protein_delins(egfr_context):
    """Create test fixture for protein delins."""
    params = {
        "id": "normalize.variation:NP_001333827.1%3Ap.Leu747_Thr751delinsPro",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.eDMXxJw9shlSKF3znIg5abniGoyJ3GQ4",
        "variation": {
            "_id": "ga4gh:VA.eDMXxJw9shlSKF3znIg5abniGoyJ3GQ4",
            "location": {
                "_id": "ga4gh:VSL.Mm8duqYDJyel5ZnwScnxLyGH1i9lcl3T",
                "interval": {
                    "end": {"value": 751, "type": "Number"},
                    "start": {"value": 746, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.vyo55F6mA6n2LgN4cagcdRzOuh38V4mE",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "P",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "LREAT",
        "gene_context": egfr_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def coding_dna_deletion(erbb2_context):
    """Create test fixture for coding dna deletion range with deleted
    sequence.
    """
    params = {
        "id": "normalize.variation:NM_004448.3%3Ac.2264_2278delTGAGGGAAAACACAT",  # noqa: E501
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.NUCURWYivhjC4oyBtzgJZ27SaaMY08Q7",
        "variation": {
            "_id": "ga4gh:VA.NUCURWYivhjC4oyBtzgJZ27SaaMY08Q7",
            "location": {
                "_id": "ga4gh:VSL.3uPWAjsdzd8MbAqw8DV46eBLK8tQRyEs",
                "interval": {
                    "end": {"value": 2453, "type": "Number"},
                    "start": {"value": 2437, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.y9b4LVMiCXpZxOg9Xt1NwRtssA03MwWM",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "T",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:0000159",
        "vrs_ref_allele_seq": "TTGAGGGAAAACACAT",
        "gene_context": erbb2_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def coding_dna_insertion(limk2_gene_context):
    """Create test fixture for coding DNA insertion."""
    params = {
        "id": "normalize.variation:ENST00000331728.9%3Ac.2049_2050insA",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.RAcEF24tRNB_J5Hz9E6GY-qQSi4ZG932",
        "variation": {
            "_id": "ga4gh:VA.RAcEF24tRNB_J5Hz9E6GY-qQSi4ZG932",
            "location": {
                "_id": "ga4gh:VSL.WKymELRFwu4LDDmH5ci5Ip0M3XA2RObr",
                "interval": {
                    "end": {"value": 2160, "type": "Number"},
                    "start": {"value": 2160, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.7_mlQyDN-uWH0RlxTQFvFEv6ykd2D-xF",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "A",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:0000667",
        "gene_context": limk2_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def genomic_insertion(erbb2_context):
    """Create a gene insertion test fixture."""
    params = {
        "id": "normalize.variation:NC_000017.10%3Ag.37880993_37880994insGCTTACGTGATG",  # noqa: E501
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.nHB0_mpsq2t90S-znr81oCi2cY5CMdUe",
        "variation": {
            "_id": "ga4gh:VA.nHB0_mpsq2t90S-znr81oCi2cY5CMdUe",
            "location": {
                "_id": "ga4gh:VSL.E0o4HCXjy1EUthF1m32oj_Bc45g5YmEm",
                "interval": {
                    "end": {"value": 2500, "type": "Number"},
                    "start": {"value": 2488, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.y9b4LVMiCXpZxOg9Xt1NwRtssA03MwWM",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "TACGTGATGGCTTACGTGATGGCT",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:0000667",
        "vrs_ref_allele_seq": "TACGTGATGGCT",
        "gene_context": erbb2_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def genomic_substitution(egfr_context):
    """Create a gene insertion test fixture."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.55249071C%3ET",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.VkcuqgqMuSQeq8Hy0VPOGRIeyr8uSBV2",
        "variation": {
            "_id": "ga4gh:VA.VkcuqgqMuSQeq8Hy0VPOGRIeyr8uSBV2",
            "location": {
                "_id": "ga4gh:VSL.G1gIZ-om-8Exl3F0ZLxXYY8CjliwCaO1",
                "interval": {
                    "end": {"value": 2630, "type": "Number"},
                    "start": {"value": 2629, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.d_QsP29RWJi6bac7GOC9cJ9AO7s_HUMN",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "T",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "transcript",
        "structural_type": "SO:0001483",
        "vrs_ref_allele_seq": "C",
        "gene_context": egfr_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def genomic_sub_grch38():
    """Create a genomic substitution GRCh38 test fixture."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.55249071C%3ET",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.1ewlywoD423K7YH_K4YefZg6J_87pQTp",
        "variation": {
            "_id": "ga4gh:VA.1ewlywoD423K7YH_K4YefZg6J_87pQTp",
            "location": {
                "_id": "ga4gh:VSL.0p1nWj9-sryfUD5jvPTZZdnZeiHVHXls",
                "interval": {
                    "end": {"value": 55181378, "type": "Number"},
                    "start": {"value": 55181377, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "T",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:0001483",
        "vrs_ref_allele_seq": "C"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def egfr_grch38_sub(genomic_sub_grch38, egfr_context):
    """Create a genomic substitution GRCh38 test fixture."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.55249071C%3ET",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.1ewlywoD423K7YH_K4YefZg6J_87pQTp",
        "variation": {
            "_id": "ga4gh:VA.1ewlywoD423K7YH_K4YefZg6J_87pQTp",
            "location": {
                "_id": "ga4gh:VSL.0p1nWj9-sryfUD5jvPTZZdnZeiHVHXls",
                "interval": {
                    "end": {"value": 55181378, "type": "Number"},
                    "start": {"value": 55181377, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "T",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:0001483",
        "vrs_ref_allele_seq": "C",
        "gene_context": egfr_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def genomic_uncertain_del_x():
    """Create a genomic uncertain deletion on chr X test fixture."""
    params = {
        "id": "normalize.variation:NC_000023.11%3Ag.%28%3F_31120496%29_%2833339477_%3F%29del",  # noqa: E501
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VAC.oCrOJZcSJ-knvR1l6u-wpPNNMCLDCB0L",
        "variation": {
            "_id": "ga4gh:VAC.oCrOJZcSJ-knvR1l6u-wpPNNMCLDCB0L",
            "subject": {
                "location": {
                    "_id": "ga4gh:VSL.7OJ5EFgu_2C4zPFDUBgn-ziE6BZwsRcv",
                    "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
                    "interval": {
                        "start": {
                            "value": 31120495,
                            "comparator": "<=",
                            "type": "IndefiniteRange"
                        },
                        "end": {
                            "value": 33339477,
                            "comparator": ">=",
                            "type": "IndefiniteRange"
                        },
                        "type": "SequenceInterval"
                    },
                    "type": "SequenceLocation"
                },
                "reverse_complement": False,
                "type": "DerivedSequenceExpression"
            },
            "copies": {
                "min": 0,
                "max": 1,
                "type": "DefiniteRange"
            },
            "type": "AbsoluteCopyNumber"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:0001743"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def grch38_braf_genom_sub(braf_v600e_genomic_sub):
    """Create a genomic substitution GRCh38 test fixture for BRAF."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.140453136A%3ET",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.fZiBjQEolbkL0AxjoTZf4SOkFy9J0ebU",
        "variation": braf_v600e_genomic_sub,
        "molecule_context": "genomic",
        "structural_type": "SO:0001483",
        "vrs_ref_allele_seq": "A"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def grch38_braf_genom_silent_mutation():
    """Create a genomic silent mutation GRCh38 test fixture for BRAF."""
    params = {
        "id": "normalize.variation:NC_000007.13%3Ag.140453136%3D",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.aMwnr5rEbtPQe5gXDDO2gZO_zSqN2RmH",
        "variation": {
            "_id": "ga4gh:VA.aMwnr5rEbtPQe5gXDDO2gZO_zSqN2RmH",
            "location": {
                "_id": "ga4gh:VSL.zga82-TpYiNmBESCfvDvAz9DyvJF98I-",
                "interval": {
                    "end": {"value": 140753336, "type": "Number"},
                    "start": {"value": 140753335, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "A",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:0002073",
        "vrs_ref_allele_seq": "A"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def grch38_genomic_delins1():
    """Create a test fixture for NC_000007.13:g.140453135_140453136delinsAT."""
    params = {
        "id":
            "normalize.variation:NC_000007.13%3Ag.140453135_140453136delinsAT",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.mlJVnI7js6Tsb2GSLFlNRbCKE9zFRX5p",
        "variation": {
            "_id": "ga4gh:VA.mlJVnI7js6Tsb2GSLFlNRbCKE9zFRX5p",
            "location": {
                "_id": "ga4gh:VSL.b0Ldj2KcT2k0n0PZfqHCBH1YzQZYceYX",
                "interval": {
                    "end": {"value": 140753336, "type": "Number"},
                    "start": {"value": 140753334, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "AT",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "CA"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def grch38_genomic_delins2():
    """Create a test fixture for NC_000003.12:g.10149938delinsAA."""
    params = {
        "id": "normalize.variation:NC_000003.12%3Ag.10149938delinsAA",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.ndCPwzek-KU626kK28bQd1gWAAk2ELze",
        "variation": {
            "_id": "ga4gh:VA.ndCPwzek-KU626kK28bQd1gWAAk2ELze",
            "location": {
                "_id": "ga4gh:VSL.mVulVOKoX2frLH1XTIJfpvJa6RGdOVu_",
                "interval": {
                    "start": {"value": 10149937, "type": "Number"},
                    "end": {"value": 10149938, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "AA",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:1000032",
        "vrs_ref_allele_seq": "C"
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="module")
def grch38_genomic_insertion():
    """Create a test fixture for
    NC_000017.10:g.37880993_37880994insGCTTACGTGATG.
    """
    params = {
        "id": "normalize.variation:NC_000017.10%3Ag.37880993_37880994insGCTTACGTGATG",  # noqa: E501
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.tCjV190dUsV7tSjdR8qOLSQIR7Hr8VMe",
        "variation": {
            "_id": "ga4gh:VA.tCjV190dUsV7tSjdR8qOLSQIR7Hr8VMe",
            "location": {
                "_id": "ga4gh:VSL.fJ80Ab9JP0GXtDNeEaoDxE35tlI-k9Cd",
                "interval": {
                    "end": {"value": 39724743, "type": "Number"},
                    "start": {"value": 39724731, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.dLZ15tNO1Ur0IcGjwc3Sdi_0A6Yf4zm7",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "TACGTGATGGCTTACGTGATGGCT",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "genomic",
        "structural_type": "SO:0000667",
        "vrs_ref_allele_seq": "TACGTGATGGCT"
    }
    return VariationDescriptor(**params)


@pytest.mark.asyncio
async def test_protein_substitution(test_query_handler, braf_v600e, dis3_p63a):
    """Test that protein substitutions normalize correctly."""
    resp = await test_query_handler.normalize("     BRAF      V600E    ")
    assertion_checks(resp, braf_v600e, "BRAF      V600E")

    braf_id = "normalize.variation:BRAF%20V600E"

    resp = await test_query_handler.normalize("NP_004324.2:p.Val600Glu")
    assert resp.id == "normalize.variation:NP_004324.2%3Ap.Val600Glu"
    resp.id = braf_id
    assertion_checks(resp, braf_v600e, "NP_004324.2:p.Val600Glu")

    resp = await test_query_handler.normalize("braf v512e")
    assert resp.id == "normalize.variation:braf%20v512e"
    resp.id = braf_id
    assertion_checks(resp, braf_v600e, "braf v512e")

    resp = await test_query_handler.normalize(" NP_001365404.1:p.Val512Glu  ")
    assert resp.id == "normalize.variation:NP_001365404.1%3Ap.Val512Glu"
    resp.id = braf_id
    assertion_checks(resp, braf_v600e, "NP_001365404.1:p.Val512Glu")

    resp = await test_query_handler.normalize("DIS3 P63A")
    assertion_checks(resp, dis3_p63a, "DIS3 P63A")


@pytest.mark.asyncio
async def test_polypeptide_truncation(test_query_handler, vhl):
    """Test that polypeptide truncations normalize correctly."""
    resp = await test_query_handler.normalize("NP_000542.1:p.Tyr185Ter")
    assertion_checks(resp, vhl, "NP_000542.1:p.Tyr185Ter")


@pytest.mark.asyncio
async def test_silent_mutation(test_query_handler, vhl_silent):
    """Test that silent mutations normalize correctly."""
    resp = await test_query_handler.normalize("NP_000542.1:p.Pro61=")
    assertion_checks(resp, vhl_silent, "NP_000542.1:p.Pro61=")


@pytest.mark.asyncio
async def test_coding_dna_and_genomic_substitution(
        test_query_handler, braf_v600e_nucleotide, genomic_substitution,
        genomic_sub_grch38, egfr_grch38_sub, grch38_braf_genom_sub):
    """Test that coding dna and genomic substitutions normalize correctly."""
    resp = await test_query_handler.normalize("NM_004333.4:c.1799T>A")
    assertion_checks(resp, braf_v600e_nucleotide, "NM_004333.4:c.1799T>A")

    # MANE transcript
    refseq_id = "normalize.variation:NM_004333.4%3Ac.1799T%3EA"

    # TODO: Check if this should return a different VRS object?
    resp = await test_query_handler.normalize("ENST00000288602.10:c.1799T>A")
    assert resp.id == "normalize.variation:ENST00000288602.10%3Ac.1799T%3EA"
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "ENST00000288602.10:c.1799T>A")

    resp = await test_query_handler.normalize("BRAF V600E c.1799T>A")
    assert resp.id == "normalize.variation:BRAF%20V600E%20c.1799T%3EA"
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "BRAF V600E c.1799T>A")

    resp = await test_query_handler.normalize("BRAF V600E (c.1799T>A)")
    assert resp.id == "normalize.variation:BRAF%20V600E%20%28c.1799T%3EA%29"
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "BRAF V600E (c.1799T>A)")

    resp = await test_query_handler.normalize("BRAF c.1799T>A")
    assert resp.id == "normalize.variation:BRAF%20c.1799T%3EA"
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "BRAF c.1799T>A")

    resp = await test_query_handler.normalize("NC_000007.13:g.140453136A>T")
    assertion_checks(resp, grch38_braf_genom_sub, "NC_000007.13:g.140453136A>T")

    fixture_id = "normalize.variation:NC_000007.13%3Ag.140453136A%3ET"
    resp = await test_query_handler.normalize("7-140453136-A-T")  # 37
    assert resp.id == "normalize.variation:7-140453136-A-T"
    resp.id = fixture_id
    assertion_checks(resp, grch38_braf_genom_sub, "7-140453136-A-T")

    resp = await test_query_handler.normalize("7-140753336-A-T")  # 38
    assert resp.id == "normalize.variation:7-140753336-A-T"
    resp.id = fixture_id
    assertion_checks(resp, grch38_braf_genom_sub, "7-140753336-A-T")

    resp = await test_query_handler.normalize("BRAF V600E (g.140453136A>T)")
    assert resp.id == "normalize.variation:BRAF%20V600E%20%28g.140453136A%3ET%29"  # noqa: E501
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "BRAF V600E (g.140453136A>T)")

    resp = await test_query_handler.normalize("BRAF g.140453136A>T")
    assert resp.id == "normalize.variation:BRAF%20g.140453136A%3ET"
    resp.id = refseq_id
    assertion_checks(resp, braf_v600e_nucleotide, "BRAF g.140453136A>T")

    # More than 1 gene (EGFR and EGFR-AS1)
    resp = await test_query_handler.normalize("NC_000007.13:g.55249071C>T")
    assertion_checks(resp, genomic_sub_grch38, "NC_000007.13:g.55249071C>T")

    resp = await test_query_handler.normalize("EGFR g.55249071C>T")
    assert resp.id == "normalize.variation:EGFR%20g.55249071C%3ET"
    resp.id = "normalize.variation:NC_000007.13%3Ag.55249071C%3ET"
    assertion_checks(resp, genomic_substitution, "EGFR g.55249071C>T")


@pytest.mark.asyncio
async def test_coding_dna_silent_mutation(test_query_handler,
                                          coding_dna_silent_mutation,
                                          braf_gene_context):
    """Test that Coding DNA Silent Mutation normalizes correctly."""
    resp = await test_query_handler.normalize("NM_004333.4:c.1799= ")
    assertion_checks(resp, coding_dna_silent_mutation, "NM_004333.4:c.1799=")

    fixture_id = "normalize.variation:NM_004333.4%3Ac.1799%3D"

    resp = await test_query_handler.normalize("ENST00000288602.11:c.1799=")
    assert resp.id == "normalize.variation:ENST00000288602.11%3Ac.1799%3D"
    resp.id = fixture_id
    assertion_checks(resp, coding_dna_silent_mutation, "ENST00000288602.11:c.1799=")

    # TODO: What to do for older Ensembl transcripts that aren"t found
    #  in seqrepo or UTA
    # resp = await test_query_handler.normalize("ENST00000288602.6:c.1799=")
    # assert_coding_dna_genomic_silent_mutation(resp, braf_gene_context,
    #                                           1798, 1799)
    # assert resp.id == "normalize.variation:ENST00000288602.6%3Ac.1799%3D"
    # assert resp.label == "ENST00000288602.6:c.1799="
    # assert resp.molecule_context == "transcript"

    resp = await test_query_handler.normalize("BRAF    c.1799=")
    assert resp.id == "normalize.variation:BRAF%20c.1799%3D"
    resp.id = fixture_id
    assertion_checks(resp, coding_dna_silent_mutation, "BRAF    c.1799=")

    resp = await test_query_handler.normalize("  BRAF  V600E  c.1799=  ")
    assert resp.id == "normalize.variation:BRAF%20V600E%20c.1799%3D"
    resp.id = fixture_id
    assertion_checks(resp, coding_dna_silent_mutation, "BRAF  V600E  c.1799=")


@pytest.mark.asyncio
async def test_genomic_silent_mutation(test_query_handler, nc_000007_silent_mutation,
                                       braf_gene_context,
                                       grch38_braf_genom_silent_mutation):
    """Test that genomic silent mutation normalizes correctly."""
    resp = await test_query_handler.normalize("NC_000007.13:g.140453136=")
    assertion_checks(resp, grch38_braf_genom_silent_mutation,
                     "NC_000007.13:g.140453136=")

    fixture_id = "normalize.variation:NC_000007.13%3Ag.140453136%3D"
    resp = await test_query_handler.normalize("7-140453136-A-A")
    assert resp.id == "normalize.variation:7-140453136-A-A"
    resp.id = fixture_id
    assertion_checks(resp, grch38_braf_genom_silent_mutation, "7-140453136-A-A")

    resp = await test_query_handler.normalize("7-140753336-A-A")
    assert resp.id == "normalize.variation:7-140753336-A-A"
    resp.id = fixture_id
    assertion_checks(resp, grch38_braf_genom_silent_mutation, "7-140753336-A-A")

    resp = await test_query_handler.normalize("BRAF g.140453136=")
    assert resp.id == "normalize.variation:BRAF%20g.140453136%3D"
    resp.id = fixture_id
    assertion_checks(resp, nc_000007_silent_mutation, "BRAF g.140453136=")


@pytest.mark.asyncio
async def test_coding_dna_delins(test_query_handler, nm_004448_coding_dna_delins,
                                 nm_000551):
    """Test that Coding DNA DelIns normalizes correctly."""
    resp = await test_query_handler.normalize("    NM_004448.4:c.2326_2327delinsCT    ")
    assertion_checks(resp, nm_004448_coding_dna_delins,
                     "NM_004448.4:c.2326_2327delinsCT")

    # TODO: Test ENST###.c

    resp = await test_query_handler.normalize("NM_000551.3:c.615delinsAA")
    nm_000551.id = "normalize.variation:NM_000551.3%3Ac.615delinsAA"
    assertion_checks(resp, nm_000551, "NM_000551.3:c.615delinsAA")


@pytest.mark.asyncio
async def test_genomic_delins(test_query_handler, nc_000007_genomic_delins,
                              nm_000551, grch38_genomic_delins1,
                              grch38_genomic_delins2):
    """Test that Genomic DelIns normalizes correctly."""
    resp = await test_query_handler.normalize(
        "NC_000007.13:g.140453135_140453136delinsAT"
    )
    assertion_checks(resp, grch38_genomic_delins1,
                     "NC_000007.13:g.140453135_140453136delinsAT")

    resp = await test_query_handler.normalize("NC_000003.12:g.10149938delinsAA")
    assertion_checks(resp, grch38_genomic_delins2, "NC_000003.12:g.10149938delinsAA")


@pytest.mark.asyncio
async def test_protein_delins(test_query_handler, protein_delins):
    """Test that Amnio Acid DelIns normalizes correctly."""
    resp = await test_query_handler.normalize("NP_001333827.1:p.Leu747_Thr751delinsPro")
    assertion_checks(resp, protein_delins, "NP_001333827.1:p.Leu747_Thr751delinsPro")

    resp = await test_query_handler.normalize("EGFR p.Leu747_Thr751delinsPro")
    assert resp.id == "normalize.variation:EGFR%20p.Leu747_Thr751delinsPro"
    resp.id = "normalize.variation:NP_001333827.1%3Ap.Leu747_Thr751delinsPro"
    assertion_checks(resp, protein_delins, "EGFR p.Leu747_Thr751delinsPro")

    resp = await test_query_handler.normalize("EGFR Leu747_Thr751delinsPro")
    assert resp.id == "normalize.variation:EGFR%20Leu747_Thr751delinsPro"
    resp.id = "normalize.variation:NP_001333827.1%3Ap.Leu747_Thr751delinsPro"
    assertion_checks(resp, protein_delins, "EGFR Leu747_Thr751delinsPro")

    resp = await test_query_handler.normalize("EGFR L747_T751delinsP")
    assert resp.id == "normalize.variation:EGFR%20L747_T751delinsP"
    resp.id = "normalize.variation:NP_001333827.1%3Ap.Leu747_Thr751delinsPro"
    assertion_checks(resp, protein_delins, "EGFR L747_T751delinsP")


@pytest.mark.asyncio
async def test_protein_deletion(test_query_handler, protein_deletion_np_range):
    """Test that Protein Deletion normalizes correctly."""
    resp = await test_query_handler.normalize("NP_004439.2:p.Leu755_Thr759del")
    assertion_checks(resp, protein_deletion_np_range,
                     "NP_004439.2:p.Leu755_Thr759del")

    resp = await test_query_handler.normalize("ERBB2 p.Leu755_Thr759del")
    assert resp.id == "normalize.variation:ERBB2%20p.Leu755_Thr759del"
    resp.id = "normalize.variation:NP_004439.2%3Ap.Leu755_Thr759del"
    assertion_checks(resp, protein_deletion_np_range, "ERBB2 p.Leu755_Thr759del")

    resp = await test_query_handler.normalize("ERBB2 Leu755_Thr759del")
    assert resp.id == "normalize.variation:ERBB2%20Leu755_Thr759del"
    resp.id = "normalize.variation:NP_004439.2%3Ap.Leu755_Thr759del"
    assertion_checks(resp, protein_deletion_np_range, "ERBB2 Leu755_Thr759del")


@pytest.mark.asyncio
async def test_coding_dna_deletion(test_query_handler, coding_dna_deletion):
    """Test that coding dna deletion normalizes correctly."""
    # https://reg.clinicalgenome.org/redmine/projects/registry/genboree_registry/by_caid?caid=CA645372623  # noqa: E501
    q = "NM_004448.3:c.2264_2278delTGAGGGAAAACACAT"
    resp = await test_query_handler.normalize(q)
    assertion_checks(resp, coding_dna_deletion, q)

    q = "ERBB2 c.2264_2278delTGAGGGAAAACACAT"
    resp = await test_query_handler.normalize(q)
    assert resp.id == \
           "normalize.variation:ERBB2%20c.2264_2278delTGAGGGAAAACACAT"
    resp.id = "normalize.variation:NM_004448.3%3Ac.2264_2278delTGAGGGAAAACACAT"
    assertion_checks(resp, coding_dna_deletion, q)


@pytest.mark.asyncio
async def test_protein_insertion(test_query_handler, protein_insertion):
    """Test that protein insertion normalizes correctly."""
    resp = await test_query_handler.normalize("NP_005219.2:p.Asp770_Asn771insGlyLeu")
    assertion_checks(resp, protein_insertion, "NP_005219.2:p.Asp770_Asn771insGlyLeu")

    def change_resp(response):
        fixture_id = \
            "normalize.variation:NP_005219.2%3Ap.Asp770_Asn771insGlyLeu"
        response.id = fixture_id

    resp = await test_query_handler.normalize("EGFR D770_N771insGL")
    assert resp.id == "normalize.variation:EGFR%20D770_N771insGL"
    change_resp(resp)
    assertion_checks(resp, protein_insertion, "EGFR D770_N771insGL")

    resp = await test_query_handler.normalize("EGFR p.D770_N771insGL")
    assert resp.id == "normalize.variation:EGFR%20p.D770_N771insGL"
    change_resp(resp)
    assertion_checks(resp, protein_insertion, "EGFR p.D770_N771insGL")

    resp = await test_query_handler.normalize("EGFR Asp770_Asn771insGlyLeu")
    assert resp.id == "normalize.variation:EGFR%20Asp770_Asn771insGlyLeu"
    change_resp(resp)
    assertion_checks(resp, protein_insertion, "EGFR Asp770_Asn771insGlyLeu")

    resp = await test_query_handler.normalize("EGFR p.Asp770_Asn771insGlyLeu")
    assert resp.id == "normalize.variation:EGFR%20p.Asp770_Asn771insGlyLeu"
    change_resp(resp)
    assertion_checks(resp, protein_insertion, "EGFR p.Asp770_Asn771insGlyLeu")


@pytest.mark.asyncio
async def test_coding_dna_insertion(test_query_handler, coding_dna_insertion):
    """Test that coding dna insertion normalizes correctly."""
    resp = await test_query_handler.normalize("ENST00000331728.9:c.2049_2050insA")
    assertion_checks(resp, coding_dna_insertion, "ENST00000331728.9:c.2049_2050insA")

    # TODO: issue-136
    # resp = await test_query_handler.normalize("LIMK2 c.2049_2050insA")
    # assert resp.id == "normalize.variation:LIMK2%20c.2049_2050insA"
    # resp.id = "normalize.variation:ENST00000331728.9%3Ac.2049_2050insA"
    # assertion_checks(resp, coding_dna_insertion)


@pytest.mark.asyncio
async def test_genomic_insertion(test_query_handler, genomic_insertion,
                                 grch38_genomic_insertion):
    """Test that genomic insertion normalizes correctly."""
    resp = await test_query_handler.normalize("NC_000017.10:g.37880993_37880994insGCTTACGTGATG")  # noqa: E501
    assertion_checks(resp, grch38_genomic_insertion,
                     "NC_000017.10:g.37880993_37880994insGCTTACGTGATG")

    fixture_id = \
        "normalize.variation:NC_000017.10%3Ag.37880993_37880994insGCTTACGTGATG"
    resp = await test_query_handler.normalize("17-37880993-G-GGCTTACGTGATG")
    assert resp.id == "normalize.variation:17-37880993-G-GGCTTACGTGATG"
    resp.id = fixture_id
    assertion_checks(resp, grch38_genomic_insertion, "17-37880993-G-GGCTTACGTGATG")

    resp = await test_query_handler.normalize(
        "ERBB2 g.37880993_37880994insGCTTACGTGATG")
    assert resp.id ==\
           "normalize.variation:ERBB2%20g.37880993_37880994insGCTTACGTGATG"
    resp.id = fixture_id
    assertion_checks(resp, genomic_insertion,
                     "ERBB2 g.37880993_37880994insGCTTACGTGATG")


@pytest.mark.asyncio
async def test_valid_queries(test_query_handler):
    """Test that valid queries don"t throw exceptions. Used for queries that
    revealed bugs in service.
    """
    assert await test_query_handler.normalize("CCND1 Y44D")

    resp = await test_query_handler.normalize("NC_000002.12:g.73448098_73448100delCTC")
    assert resp
    assert resp.variation.state.sequence == "CTC"
    assert resp.variation.id == "ga4gh:VA.7qNgfRjiDwephtGOuwtXm04YcXDSY47x"


@pytest.mark.asyncio
async def test_no_matches(test_query_handler):
    """Test no matches work correctly."""
    queries = [
        "braf", "braf v600000932092039e", "NP_000213.1:cp.Leu862=",
        "NP_000213.1:cp.Leu862", "BRAF V600E 33", "NP_004324.2:p.Glu600Val",
        "NP_004324.2:p.Glu600Gal", "NP_004324.2839:p.Glu600Val",
        "NP_004324.2:t.Glu600Val", "this:c.54G>H", "NC_000007.13:g.4T<A",
        "test", "131", "braf z600e", "braf e600z", "Thr790Met", "p.Tyr365Ter",
        "ERBB2 G776delinsVCZ", "NP005219.2:p.Glu746_Thr751delinsValAla",
        "NP_005219.2:p.Glu746Thr751delinsValAla", "EGFR L747_L474delinsP",
        "NP_005219.2:p.Glu746_Thr751delinssValAla", "EGFR delins",
        "NM_004333.4:c.1799_1800delTGinsAT",
        "NM_173851.3(SLC30A8):c.973C>T%20(p.Arg325Trp)"
    ]
    for q in queries:
        resp = await test_query_handler.normalize(q)
        assert resp.type == "VariationDescriptor", q
        assert resp.variation.type == "Text", q
        assert resp.label == q.strip(), q

    resp = await test_query_handler.normalize("clinvar:10")
    assert resp.type == "VariationDescriptor"
    assert resp.label == "clinvar:10"
    assert resp.variation.type == "Text"
    assert resp.variation.definition == "clinvar:10"
    assert resp.variation.id == "ga4gh:VT.xw9m9LZAyn6Z2-GPGwcpDT0ixqCm5g36"

    resp = await test_query_handler.normalize("   ")
    assert resp is None

    resp = await test_query_handler.normalize("")
    assert resp is None

    resp = await test_query_handler.normalize(None)
    assert resp is None


@pytest.mark.asyncio
async def test_service_meta():
    """Test that service meta info populates correctly."""
    response = await normalize_get_response("BRAF v600e", "default")
    service_meta = response.service_meta_
    assert service_meta.name == "variation-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/variation-normalization"  # noqa: E501

    response = await normalize_get_response("this-wont-normalize", "default")
    service_meta = response.service_meta_
    assert service_meta.name == "variation-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/variation-normalization"  # noqa: E501

    response = await to_vrs_get_response("BRAF v600e")
    service_meta = response.service_meta_
    assert service_meta.name == "variation-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/variation-normalization"  # noqa: E501

    response = await to_vrs_get_response("this-wont-normalize")
    service_meta = response.service_meta_
    assert service_meta.name == "variation-normalizer"
    assert service_meta.version
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/variation-normalization"  # noqa: E501



"""A module for Protein Deletion Tokenization Class."""
from typing import Optional, List

from pydantic.error_wrappers import ValidationError

from variation.schemas.token_response_schema import ProteinDeletionToken, \
    TokenMatchType
from .caches import AminoAcidCache, NucleotideCache
from .tokenizer import Tokenizer
from .tokenize_base import TokenizeBase


class ProteinDeletion(Tokenizer):
    """Class for tokenizing Deletions on the protein reference sequence."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 nucleotide_cache: NucleotideCache) -> None:
        """Initialize the Protein Deletion Class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes
        :param NucleotideCache nucleotide_cache: Valid nucleotides
        """
        self.parts = None
        self.tokenize_base = TokenizeBase(amino_acid_cache, nucleotide_cache)

    def match(self, input_string: str) -> Optional[ProteinDeletionToken]:
        """Return token that match the input string."""
        if input_string is None:
            return None

        self.parts = {
            "used_one_letter": False,
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_aa_del": None,
            "start_pos_del": None,
            "end_aa_del": None,
            "end_pos_del": None
        }

        input_string = str(input_string).lower()

        if "c." in input_string or "g." in input_string:
            return None

        if input_string.startswith("p."):
            input_string = input_string[2:]

        if input_string.startswith("(") and input_string.endswith(")"):
            input_string = input_string[1:-1]

        if not input_string.endswith("del"):
            return None

        parts = input_string.split("del")
        self._get_parts(parts)

        try:
            return ProteinDeletionToken(**self.parts)
        except ValidationError:
            return None

    def _get_parts(self, parts: List) -> None:
        """Get parts for Protein Deletion.

        :param List parts: Parts of input string
        """
        if len(parts) != 2:
            return

        range_aa_pos = self.tokenize_base.get_aa_pos_range(parts)
        if range_aa_pos:
            self.parts["start_aa_del"] = range_aa_pos[0]
            self.parts["end_aa_del"] = range_aa_pos[1]
            self.parts["start_pos_del"] = range_aa_pos[2]
            self.parts["end_pos_del"] = range_aa_pos[3]
            self.parts["used_one_letter"] = range_aa_pos[4]



"""Module for testing the hgvs to absolute and relative copy number endpoints"""
import copy

import pytest


@pytest.fixture(scope="module")
def genomic_dup1_38_subject(genomic_dup1_seq_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genomic_dup1_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup1_rel_38(genomic_dup1_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.XiXamTGYJ43rc8xheleMKcjxEBOFp82l",
        "subject": genomic_dup1_38_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_dup1_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.CXcLL6RUPkro3dLXN0miGEzlzPYiqw2q",
            "sequence_id": "ga4gh:SQ.VNBualIltAyi2AI_uXcKU7M9XUOuA7MS",
            "interval": {
                "type": "SequenceInterval",
                "start": {"value": 49568693, "type": "Number"},
                "end": {"value": 49568695, "type": "Number"},
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup1_abs_37(genomic_dup1_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.ReDrt_J0qRXOokz_32CWYdaWy7aATEHT",
        "subject": genomic_dup1_37_subject,
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup1_rel_37(genomic_dup1_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.Y7OyCA7q7mil6YFiHEh4pyZHPLkGaPDE",
        "subject": genomic_dup1_37_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_dup2_38_subject(genomic_dup2_seq_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genomic_dup2_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup2_rel_38(genomic_dup2_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.ZafHwmugzRCgqgy4jAS6W-l3y3gSEezQ",
        "subject": genomic_dup2_38_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_dup2_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.VY7qo3C2Nl-bpRtjOQ6MODJ2FfGUJRV1",
            "sequence_id": "ga4gh:SQ.W6wLoIFOn4G7cjopxPxYNk2lcEqhLQFb",
            "interval": {
                "type": "SequenceInterval",
                "start": {"value": 2137938, "type": "Number"},
                "end": {"value": 2137949, "type": "Number"},
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup2_abs_37(genomic_dup2_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.8N1BeYiz3saze_AC26Gg4ocKF8IO7GqF",
        "subject": genomic_dup2_37_subject,
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup2_rel_37(genomic_dup2_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.1vUSqjdiWCEJrPYfGs_4PDEVLHohztdJ",
        "subject": genomic_dup2_37_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_dup3_38_subject(genomic_del3_dup3_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genomic_del3_dup3_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup3_abs_38(genomic_dup3_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.LT_jOeYNJg6UblsnItIhX0tppRff4fCh",
        "subject": genomic_dup3_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup3_rel_38(genomic_dup3_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.-y-_z5rsOV0x-KsgPjYiDN6XDNV4MX7U",
        "subject": genomic_dup3_38_subject,
        "relative_copy_class": "high-level gain"
    }


@pytest.fixture(scope="module")
def genomic_dup3_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.hDvTkhN210ZKyZOhilS7FxaKUCJgj6JC",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "min": 31078343,
                    "max": 31118467,
                    "type": "DefiniteRange"
                },
                "end": {
                    "min": 33292396,
                    "max": 33435269,
                    "type": "DefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup3_abs_37(genomic_dup3_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.LAY9brjDTRS9v6EHIqorBJSWdPCdGehM",
        "subject": genomic_dup3_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup3_rel_37(genomic_dup3_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.sPsCwxoV3zg4laO26t0iYm_FQoTThCBF",
        "subject": genomic_dup3_37_subject,
        "relative_copy_class": "high-level gain"
    }


@pytest.fixture(scope="module")
def genomic_dup4_38_subject(genoimc_dup4_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genoimc_dup4_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup4_abs_38(genomic_dup4_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.aHxuxlXZTVm-HroDd34Jh9BNPrgVHtML",
        "subject": genomic_dup4_38_subject,
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup4_rel_38(genomic_dup4_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.E2IxQfOBiCWvuXa7U-swwgbRoiV5aURe",
        "subject": genomic_dup4_38_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_dup4_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.ypfdWazLg2-FkZhRH0EQDGwH5IqlAyLh",
            "sequence_id": "ga4gh:SQ.iy_UbUrvECxFRX5LPTH_KPojdlT7BKsf",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 29652251,
                    "comparator": "<=",
                    "type": "IndefiniteRange"
                },
                "end": {
                    "value": 29981821,
                    "comparator": ">=",
                    "type": "IndefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup4_abs_37(genomic_dup4_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.2NR9ZYtmbFy31pIic_1cBZ9SEev0q-Sx",
        "subject": genomic_dup4_37_subject,
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup4_rel_37(genomic_dup4_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.JAk-UphGbw9-rFi0WoHJNW-CZKBIA81I",
        "subject": genomic_dup4_37_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_dup5_38_subject(genoimc_dup5_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genoimc_dup5_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup5_abs_38(genomic_dup5_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.aaz7BxN1V9TzvG934_kSQbbPNCT6vHU1",
        "subject": genomic_dup5_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup5_rel_38(genomic_dup5_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.dfAwyWhv8FlJLXGG8fS86qp5kSmgK9tb",
        "subject": genomic_dup5_38_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_dup5_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.5pVsde3FHXnOaO0jONcisMHAfcjqmXiD",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 153287262,
                    "comparator": "<=",
                    "type": "IndefiniteRange"
                },
                "end": {
                    "value": 153357667,
                    "type": "Number"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup5_abs_37(genomic_dup5_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.99BliOHLTPSR7gqbPNKTCfRMXHFrwqgd",
        "subject": genomic_dup5_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup5_rel_37(genomic_dup5_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.u7y0KCsAhXEy3bdcfYwOtGBBGNziLb3W",
        "subject": genomic_dup5_37_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_dup6_38_subject(genoimc_dup6_loc):
    """Create test fixture GRCh38 duplication subject"""
    return {
        "location": genoimc_dup6_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup6_abs_38(genomic_dup6_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.v-dhCkHL7W67y3SHp-qfMg7F5DOaAbQn",
        "subject": genomic_dup6_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup6_rel_38(genomic_dup6_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.dLmIPRSPDfaZsY6JkQZJPumqsZTHgvb8",
        "subject": genomic_dup6_38_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_dup6_37_subject():
    """Create test fixture GRCh37 duplication subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.4FthtlBwSSEa3O0CUPw-DnKYz4Jnfa3n",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 153287262,
                    "type": "Number"
                },
                "end": {
                    "value": 153357667,
                    "comparator": ">=",
                    "type": "IndefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_dup6_abs_37(genomic_dup6_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.aIHl5yt6YJPqrbh4FL6Fmlg7DTEqCh8y",
        "subject": genomic_dup6_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 2,
            "max": 3
        }
    }


@pytest.fixture(scope="module")
def genomic_dup6_rel_37(genomic_dup6_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.LkA6PMk1quZ0K_bgGhI6XpmCylDocIR_",
        "subject": genomic_dup6_37_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_del1_38_subject(genomic_del1_seq_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del1_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del1_rel_38(genomic_del1_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.KQwe4TvYODBZeSUqC7K3qssIRQhxnevQ",
        "subject": genomic_del1_38_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_del1_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.IpNE_OIaSJSeiiJTEwrxyiRsgpUay8_n",
            "sequence_id": "ga4gh:SQ.VNBualIltAyi2AI_uXcKU7M9XUOuA7MS",
            "interval": {
                "type": "SequenceInterval",
                "start": {"value": 10191494, "type": "Number"},
                "end": {"value": 10191495, "type": "Number"},
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del1_abs_37(genomic_del1_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.BJcNDL3rGhci1ftMCL0_exxDb45SWgxn",
        "subject": genomic_del1_37_subject,
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del1_rel_37(genomic_del1_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.T-BAA72t1pfMsgxCaJGfS8vg_cdS-8dX",
        "subject": genomic_del1_37_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_del2_38_subject(genomic_del2_seq_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del2_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del2_rel_38(genomic_del2_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.DRcKvwgwmtxjpRUctWaFBgzK-3mu7p7L",
        "subject": genomic_del2_38_subject,
        "relative_copy_class": "low-level gain"
    }


@pytest.fixture(scope="module")
def genomic_del2_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.B-feFlnwavUnci2jFAqG8WOQrqFAZJ6N",
            "sequence_id": "ga4gh:SQ.VNBualIltAyi2AI_uXcKU7M9XUOuA7MS",
            "interval": {
                "type": "SequenceInterval",
                "start": {"value": 10188278, "type": "Number"},
                "end": {"value": 10188297, "type": "Number"},
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del2_abs_37(genomic_del2_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.PgSgggCYoZfyVIDT5LIYemgu6UQHAFOh",
        "subject": genomic_del2_37_subject,
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del2_rel_37(genomic_del2_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.yV0_s897kIPggWedY4buNUzaokXTVzZG",
        "subject": genomic_del2_37_subject,
        "relative_copy_class": "low-level gain"
    }


@pytest.fixture(scope="module")
def genomic_del3_38_subject(genomic_del3_dup3_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del3_dup3_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del3_abs_38(genomic_del3_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.o39DwfEb2I-LR9NFKR5cznWrRhI9Vx8b",
        "subject": genomic_del3_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del3_rel_38(genomic_del3_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.vpFW_FTFwvziXVlkOsxZTTFn0rvzHfnA",
        "subject": genomic_del3_38_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_del3_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.hDvTkhN210ZKyZOhilS7FxaKUCJgj6JC",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "min": 31078343,
                    "max": 31118467,
                    "type": "DefiniteRange"
                },
                "end": {
                    "min": 33292396,
                    "max": 33435269,
                    "type": "DefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del3_abs_37(genomic_del3_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.0EC3Bi-mP2GtfIOhPMvYjhXVHIUhxYQP",
        "subject": genomic_del3_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del3_rel_37(genomic_del3_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.G1O15kvCHpTTTyAeuWNFG0rTj1SgvzFq",
        "subject": genomic_del3_37_subject,
        "relative_copy_class": "complete loss"
    }


@pytest.fixture(scope="module")
def genomic_del4_38_subject(genomic_del4_seq_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del4_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del4_abs_38(genomic_del4_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.oCrOJZcSJ-knvR1l6u-wpPNNMCLDCB0L",
        "subject": genomic_del4_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del4_rel_38(genomic_del4_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.u7we9nJeMlt-CAlQ47HRN7lWJ0UXay_W",
        "subject": genomic_del4_38_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_del4_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.7Chaj1X9NH2G9sSK1diUKgBEZO4pHqr8",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 31138612,
                    "comparator": "<=",
                    "type": "IndefiniteRange"
                },
                "end": {
                    "value": 33357594,
                    "comparator": ">=",
                    "type": "IndefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del4_abs_37(genomic_del4_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.sBsyUfIqJcBr2Sl7rsscPcjlId-wX9_V",
        "subject": genomic_del4_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del4_rel_37(genomic_del4_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.E2X_1uJAfCD5LVHgv6kJKrq0bgUxZmEs",
        "subject": genomic_del4_37_subject,
        "relative_copy_class": "partial loss"
    }


@pytest.fixture(scope="module")
def genomic_del5_38_subject(genomic_del5_seq_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del5_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del5_abs_38(genomic_del5_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.WQkCH9rBzLPtB4f6LRb8Y-rhob68h-7M",
        "subject": genomic_del5_38_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del5_rel_38(genomic_del5_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.WNlJVKAmuN8BgZDvl584zG60asYKo7Iz",
        "subject": genomic_del5_38_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_del5_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.BXq1jg19bLLthS20Ux3MZJaXEeAAyuW9",
            "sequence_id": "ga4gh:SQ.v7noePfnNpK8ghYXEqZ9NukMXW7YeNsm",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 18593473,
                    "comparator": "<=",
                    "type": "IndefiniteRange"
                },
                "end": {
                    "value": 18671749,
                    "type": "Number"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del5_abs_37(genomic_del5_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.sA4s5CrkkknwUt64RUVNcy0AqY3uD3W4",
        "subject": genomic_del5_37_subject,
        "copies": {
            "type": "DefiniteRange",
            "min": 0,
            "max": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del5_rel_37(genomic_del5_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.waFZlE_5ZXtD02BR4bWqYJGQ7LNf6I7a",
        "subject": genomic_del5_37_subject,
        "relative_copy_class": "copy neutral"
    }


@pytest.fixture(scope="module")
def genomic_del6_38_subject(genomic_del6_seq_loc):
    """Create test fixture GRCh38 deletion subject"""
    return {
        "location": genomic_del6_seq_loc,
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del6_abs_38(genomic_del6_38_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.YAyl7c-YMRkbqZymPFI2GKkYSQ9v7EeD",
        "subject": genomic_del6_38_subject,
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del6_rel_38(genomic_del6_38_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.wS7t1-8Q0XfNKF9SbOEL_mD7X8joaslw",
        "subject": genomic_del6_38_subject,
        "relative_copy_class": "low-level gain"
    }


@pytest.fixture(scope="module")
def genomic_del6_37_subject():
    """Create test fixture GRCh37 deletion subject"""
    return {
        "location": {
            "_id": "ga4gh:VSL.FtOUv03wexaJWg_U7DISUtrM5WPdX60d",
            "sequence_id": "ga4gh:SQ.KqaUhJMW3CDjhoVtBetdEKT1n6hM-7Ek",
            "interval": {
                "type": "SequenceInterval",
                "start": {
                    "value": 133783901,
                    "type": "Number"
                },
                "end": {
                    "value": 133785996,
                    "comparator": ">=",
                    "type": "IndefiniteRange"
                }
            },
            "type": "SequenceLocation",
        },
        "reverse_complement": False,
        "type": "DerivedSequenceExpression"
    }


@pytest.fixture(scope="module")
def genomic_del6_abs_37(genomic_del6_37_subject):
    """Create test fixture absolute copy number variation"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.V9-mDixtWzx19X7dAdzTvXq4hXqQ9udx",
        "subject": genomic_del6_37_subject,
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="module")
def genomic_del6_rel_37(genomic_del6_37_subject):
    """Create test fixture relative copy number variation"""
    return {
        "type": "RelativeCopyNumber",
        "_id": "ga4gh:VRC.ah-x-uIo3eTsIMXCvmXkEvhzm3d52X7Y",
        "subject": genomic_del6_37_subject,
        "relative_copy_class": "low-level gain"
    }


@pytest.mark.asyncio
async def test_genomic_dup1_absolute_cnv(test_query_handler, genomic_dup1_38_vac,
                                         genomic_dup1_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000003.12:g.49531262dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup1_38_vac

    q = "NC_000003.11:g.49568695dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup1_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup1_38_vac

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=1, do_liftover=True)
    expected = copy.deepcopy(genomic_dup1_38_vac)
    expected["copies"]["value"] = 2
    expected["_id"] = "ga4gh:VAC.PDhQpQe82ssZiIMlghL8wDlf6xBIV_Ca"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup1_relative_cnv(test_query_handler, genomic_dup1_rel_38,
                                         genomic_dup1_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000003.12:g.49531262dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup1_rel_38

    q = "NC_000003.11:g.49568695dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup1_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup1_rel_38


@pytest.mark.asyncio
async def test_genomic_dup2_absolute_cnv(test_query_handler, genomic_dup2_38_vac,
                                         genomic_dup2_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000016.10:g.2087938_2087948dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup2_38_vac

    q = "NC_000016.9:g.2137939_2137949dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup2_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup2_38_vac

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=1, do_liftover=True)
    expected = copy.deepcopy(genomic_dup2_38_vac)
    expected["copies"]["value"] = 2
    expected["_id"] = "ga4gh:VAC.Riu3BNPMzofH0ABV0Ulr1W1hxojyl4Qk"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup2_relative_cnv(test_query_handler, genomic_dup2_rel_38,
                                         genomic_dup2_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000016.10:g.2087938_2087948dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup2_rel_38

    q = "NC_000016.9:g.2137939_2137949dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup2_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup2_rel_38


@pytest.mark.asyncio
async def test_genomic_dup3_absolute_cnv(test_query_handler, genomic_dup3_abs_38,
                                         genomic_dup3_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.(31060227_31100351)_(33274278_33417151)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup3_abs_38

    q = "NC_000023.10:g.(31078344_31118468)_(33292395_33435268)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup3_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup3_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=2, do_liftover=True)
    expected = copy.deepcopy(genomic_dup3_abs_38)
    expected["copies"] = {"value": 3, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.kuf5yPlF13yjhXKJMGc_gMiClDlvIFAY"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup3_relative_cnv(test_query_handler, genomic_dup3_rel_38,
                                         genomic_dup3_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.(31060227_31100351)_(33274278_33417151)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="high-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup3_rel_38

    q = "NC_000023.10:g.(31078344_31118468)_(33292395_33435268)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="high-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup3_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="high-level gain", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup3_rel_38


@pytest.mark.asyncio
async def test_genomic_dup4_absolute_cnv(test_query_handler, genomic_dup4_abs_38,
                                         genomic_dup4_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000020.11:g.(?_30417576)_(31394018_?)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup4_abs_38

    q = "NC_000020.10:g.(?_29652252)_(29981821_?)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup4_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup4_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=1, do_liftover=True)
    expected = copy.deepcopy(genomic_dup4_abs_38)
    expected["copies"]["value"] = 2
    expected["_id"] = "ga4gh:VAC.qAzWMQEm93Q5FOSgsf8uWOVbiB9UZVc7"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup4_relative_cnv(test_query_handler, genomic_dup4_rel_38,
                                         genomic_dup4_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000020.11:g.(?_30417576)_(31394018_?)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup4_rel_38

    q = "NC_000020.10:g.(?_29652252)_(29981821_?)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup4_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup4_rel_38


@pytest.mark.asyncio
async def test_genomic_dup5_absolute_cnv(test_query_handler, genomic_dup5_abs_38,
                                         genomic_dup5_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.(?_154021812)_154092209dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup5_abs_38

    q = "NC_000023.10:g.(?_153287263)_153357667dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup5_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup5_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=4, do_liftover=True)
    expected = copy.deepcopy(genomic_dup5_abs_38)
    expected["copies"] = {"value": 5, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.OCz5vEoeJ8rkAcJHiYiZt9yrscRYniVL"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup5_relative_cnv(test_query_handler, genomic_dup5_rel_38,
                                         genomic_dup5_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.(?_154021812)_154092209dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup5_rel_38

    q = "NC_000023.10:g.(?_153287263)_153357667dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup5_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup5_rel_38


@pytest.mark.asyncio
async def test_genomic_dup6_absolute_cnv(test_query_handler, genomic_dup6_abs_38,
                                         genomic_dup6_abs_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.154021812_(154092209_?)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup6_abs_38

    q = "NC_000023.10:g.153287263_(153357667_?)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup6_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup6_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=2, do_liftover=True)
    expected = copy.deepcopy(genomic_dup6_abs_38)
    expected["copies"] = {"value": 3, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.lfKTjaLLNWyis9O1lpn1aBfi_YSIFCz4"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_dup6_relative_cnv(test_query_handler, genomic_dup6_rel_38,
                                         genomic_dup6_rel_37):
    """Test that genomic duplication works correctly"""
    q = "NC_000023.11:g.154021812_(154092209_?)dup"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup6_rel_38

    q = "NC_000023.10:g.153287263_(153357667_?)dup"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_dup6_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_dup6_rel_38


@pytest.mark.asyncio
async def test_genomic_del1_absolute_cnv(test_query_handler, genomic_del1_38_vac,
                                         genomic_del1_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000003.12:g.10149811del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del1_38_vac

    q = "NC_000003.11:g.10191495del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del1_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del1_38_vac

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=3, do_liftover=True)
    expected = copy.deepcopy(genomic_del1_38_vac)
    expected["copies"]["value"] = 2
    expected["_id"] = "ga4gh:VAC.YwwG6b9B7LDFsKlw_t1aRilDUu-hGajk"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del1_relative_cnv(test_query_handler, genomic_del1_rel_38,
                                         genomic_del1_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000003.12:g.10149811del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del1_rel_38

    q = "NC_000003.11:g.10191495del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del1_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del1_rel_38


@pytest.mark.asyncio
async def test_genomic_del2_absolute_cnv(test_query_handler, genomic_del2_38_vac,
                                         genomic_del2_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000003.12:g.10146595_10146613del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del2_38_vac

    q = "NC_000003.11:g.10188279_10188297del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del2_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del2_38_vac

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=4, do_liftover=True)
    expected = copy.deepcopy(genomic_del2_38_vac)
    expected["copies"]["value"] = 3
    expected["_id"] = "ga4gh:VAC.lvGZeXqN0zTOD43dzLa9HmSjYz2-MBJf"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del2_relative_cnv(test_query_handler, genomic_del2_rel_38,
                                         genomic_del2_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000003.12:g.10146595_10146613del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del2_rel_38

    q = "NC_000003.11:g.10188279_10188297del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del2_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del2_rel_38


@pytest.mark.asyncio
async def test_genomic_del3_absolute_cnv(test_query_handler, genomic_del3_abs_38,
                                         genomic_del3_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(31060227_31100351)_(33274278_33417151)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del3_abs_38

    q = "NC_000023.10:g.(31078344_31118468)_(33292395_33435268)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del3_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del3_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=2, do_liftover=True)
    expected = copy.deepcopy(genomic_del3_abs_38)
    expected["copies"] = {"value": 1, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.ohgrbhR6eL_JXnXxXd-4mXTHot2HfPJQ"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del3_relative_cnv(test_query_handler, genomic_del3_rel_38,
                                         genomic_del3_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(31060227_31100351)_(33274278_33417151)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del3_rel_38

    q = "NC_000023.10:g.(31078344_31118468)_(33292395_33435268)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del3_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="complete loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del3_rel_38


@pytest.mark.asyncio
async def test_genomic_del4_absolute_cnv(test_query_handler, genomic_del4_abs_38,
                                         genomic_del4_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(?_31120496)_(33339477_?)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del4_abs_38

    q = "NC_000023.10:g.(?_31138613)_(33357594_?)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del4_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del4_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=3, do_liftover=True)
    expected = copy.deepcopy(genomic_del4_abs_38)
    expected["copies"] = {"value": 2, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.QMZBxhPgywSZ-d1je0IKs_z6g0a2iN-X"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del4_relative_cnv(test_query_handler, genomic_del4_rel_38,
                                         genomic_del4_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(?_31120496)_(33339477_?)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del4_rel_38

    q = "NC_000023.10:g.(?_31138613)_(33357594_?)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del4_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="partial loss", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del4_rel_38


@pytest.mark.asyncio
async def test_genomic_del5_absolute_cnv(test_query_handler, genomic_del5_abs_38,
                                         genomic_del5_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(?_18575354)_18653629del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del5_abs_38

    q = "NC_000023.10:g.(?_18593474)_18671749del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del5_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del5_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=2, do_liftover=True)
    expected = copy.deepcopy(genomic_del5_abs_38)
    expected["copies"] = {"value": 1, "type": "Number"}
    expected["_id"] = "ga4gh:VAC.q4lg4INA2pSXs6cu3liKnJGacKofQDoB"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del5_relative_cnv(test_query_handler, genomic_del5_rel_38,
                                         genomic_del5_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000023.11:g.(?_18575354)_18653629del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del5_rel_38

    q = "NC_000023.10:g.(?_18593474)_18671749del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del5_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="copy neutral", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del5_rel_38


@pytest.mark.asyncio
async def test_genomic_del6_absolute_cnv(test_query_handler, genomic_del6_abs_38,
                                         genomic_del6_abs_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000006.12:g.133462764_(133464858_?)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del6_abs_38

    q = "NC_000006.11:g.133783902_(133785996_?)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del6_abs_37

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=None, do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del6_abs_38

    resp, _ = await test_query_handler.hgvs_to_absolute_copy_number(
        q, baseline_copies=3, do_liftover=True)
    expected = copy.deepcopy(genomic_del6_abs_38)
    expected["copies"]["value"] = 2
    expected["_id"] = "ga4gh:VAC.--AWt37YlF0oVaExdssLIrrLl5EEh_0k"
    assert resp.dict(by_alias=True) == expected


@pytest.mark.asyncio
async def test_genomic_del6_relative_cnv(test_query_handler, genomic_del6_rel_38,
                                         genomic_del6_rel_37):
    """Test that genomic deletion works correctly"""
    q = "NC_000006.12:g.133462764_(133464858_?)del"  # 38
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del6_rel_38

    q = "NC_000006.11:g.133783902_(133785996_?)del"  # 37
    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=False)
    assert resp.dict(by_alias=True) == genomic_del6_rel_37

    resp, _ = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=True)
    assert resp.dict(by_alias=True) == genomic_del6_rel_38


@pytest.mark.asyncio
async def test_invalid_cnv_parameters(test_query_handler):
    """Check that invalid parameters return warnings"""
    q = "NC_000006.11:g.133783902_(133785996_?)del"
    resp, w = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gains", do_liftover=True)
    assert resp is None
    assert w == ["low-level gains is not a valid relative copy class: ['complete loss', "  # noqa: E501
                 "'partial loss', 'copy neutral', 'low-level gain', 'high-level gain']"]


@pytest.mark.asyncio
async def test_invalid_cnv(test_query_handler):
    """Check that invalid input return warnings"""
    q = "DAG1 g.49568695dup"
    resp, w = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=True)
    assert set(w) == {"Unable to translate DAG1 g.49568695dup to copy number variation",
                      "DAG1 g.49568695dup is not a supported HGVS genomic duplication or deletion"}  # noqa: E501
    assert resp.type == "Text"

    q = "braf v600e"
    resp, w = await test_query_handler.hgvs_to_relative_copy_number(
        q, relative_copy_class="low-level gain", do_liftover=True)
    assert set(w) == {"Unable to translate braf v600e to copy number variation",
                      "braf v600e is not a supported HGVS genomic duplication or deletion"}  # noqa: E501
    assert resp.type == "Text"



"""Module for testing Genomic Insertion Translator."""
import unittest

from variation.classifiers import GenomicInsertionClassifier
from variation.translators import GenomicInsertion
from variation.validators import GenomicInsertion as GD_V
from .translator_base import TranslatorBase


class TestGenomicInsertionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic Insertion Translator."""

    def classifier_instance(self):
        """Return genomic insertion instance."""
        return GenomicInsertionClassifier()

    def validator_instance(self):
        """Return coding DNA insertion instance."""
        return GD_V(*self.params)

    def translator_instance(self):
        """Return genomic insertion instance."""
        return GenomicInsertion()

    def fixture_name(self):
        """Return the fixture name for genomic insertion."""
        return "genomic_insertion"



"""Module for Genomic Validation methods."""
import logging
from typing import List, Optional

from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from uta_tools.data_sources import UTADatabase

from variation.schemas.classification_response_schema import Classification


logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class GenomicBase:
    """Genomic Base class for validation methods."""

    def __init__(self, dp: SeqRepoDataProxy, uta: UTADatabase) -> None:
        """Initialize the Genomic base class.

        :param SeqRepoDataProxy dp: Access to seqrepo data
        :param UTADatabase uta: Access to UTA queries
        """
        self.dp = dp
        self.uta = uta

    """The Genomic Base class."""
    async def get_nc_accessions(self, classification: Classification) -> List[str]:
        """Get NC accession for a given classification."""
        hgvs = [t.token for t in classification.all_tokens if
                t.token_type in ["HGVS", "ReferenceSequence"]]
        nc_accessions = []
        if hgvs:
            nc_accessions = [hgvs[0].split(":")[0]]
        else:
            chromosome = [t.token for t in classification.all_tokens if
                          t.token_type in ["Chromosome"]]
            if chromosome:
                chromosome = chromosome[0]
                for assesmbly in ["GRCh37", "GRCh38"]:
                    ac = self.get_nc_accession(f"{assesmbly}:{chromosome}")
                    if ac:
                        nc_accessions.append(ac)
            else:
                gene_tokens = [t.token for t in classification.all_tokens
                               if t.token_type == "GeneSymbol"]
                if gene_tokens and len(gene_tokens) == 1:
                    nc_accessions = await self.uta.get_ac_from_gene(gene_tokens[0])
        return nc_accessions

    def get_nc_accession(self, identifier: str) -> Optional[str]:
        """Given an identifier (assembly+chr), return nc accession."""
        nc_accession = None
        try:
            metadata = \
                self.dp.get_metadata(identifier)
        except KeyError:
            logger.warning("Data Proxy unable to get metadata"
                           f"for {identifier}")
        else:
            aliases = [a for a in metadata["aliases"] if
                       a.startswith("refseq:NC_")]
            if aliases:
                nc_accession = aliases[0].split(":")[-1]
        return nc_accession



"""A module for testing Coding DNA DelIns tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import CodingDNADelIns
from .tokenizer_base import TokenizerBase


class TestCodingDNADelInsTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA DelIns Tokenization."""

    def tokenizer_instance(self):
        """Return Coding DNA DelIns instance."""
        return CodingDNADelIns(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return DNA coding delins token type."""
        return "CodingDNADelIns"

    def fixture_name(self):
        """Return the fixture name for DNA coding delins."""
        return "coding_dna_delins"



"""Module for testing Genomic Silent Mutation Classifier."""
import unittest

from variation.classifiers import GenomicSilentMutationClassifier
from .classifier_base import ClassifierBase


class TestGenomicSilentMutationClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Genomic Silent Mutation Classifier."""

    def classifier_instance(self):
        """Return GenomicSilentMutationClassifier instance."""
        return GenomicSilentMutationClassifier()

    def fixture_name(self):
        """Return GenomicSilentMutationClassifier fixture name."""
        return "genomic_silent_mutation"



"""Module for translation."""
from abc import ABC, abstractmethod
from typing import Dict, Optional

from ga4gh.vrsatile.pydantic.vrs_models import Allele, AbsoluteCopyNumber, \
    RelativeCopyNumber
from pydantic.error_wrappers import ValidationError

from variation.schemas.validation_response_schema import ValidationResult
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import Token


class Translator(ABC):
    """The translation class."""

    @abstractmethod
    def can_translate(self, type: ClassificationType) -> bool:
        """Determine if it's possible to translate a classification."""
        raise NotImplementedError

    @abstractmethod
    def is_token_instance(self, token: Token) -> bool:
        """Check that the token is the correct instance for a translator."""
        raise NotImplementedError

    def translate(self, res: ValidationResult) -> Optional[Dict]:
        """Translate to VRS Variation representation."""
        instance_tokens = [t for t in res.classification.all_tokens if
                           self.is_token_instance(t)]
        len_instance_tokens = len(instance_tokens)

        variation_type = res.variation["type"]
        if len_instance_tokens > 1:
            if len_instance_tokens == 2:
                if {t.token_type for t in instance_tokens} \
                        == {"PolypeptideTruncation"}:
                    tokens = sorted([t.token.lower() for t in instance_tokens],
                                    key=len)
                    if len(tokens[1]) > len(tokens[0]):
                        t = f"{tokens[0]} ({tokens[0].replace('ter', '*')})"
                        if t.lower() == tokens[1].lower():
                            if variation_type == "Allele":
                                try:
                                    Allele(**res.variation)
                                except ValidationError:
                                    return None
                                else:
                                    return res.variation

            raise Exception(f"Should not have more than one "
                            f"{self.__class__.__name__} "
                            f"token if the result is valid")

        if variation_type == "Allele":
            if not res.variation["location"]:
                raise Exception("Cannot translate an allele with no location")
            try:
                Allele(**res.variation)
            except ValidationError:
                variation = None
            else:
                variation = res.variation
        elif variation_type == "AbsoluteCopyNumber":
            if res.variation["subject"]["type"] == "Allele":
                if not res.variation["subject"]["location"]:
                    raise Exception("Cannot translate a CNV with no location")
            try:
                AbsoluteCopyNumber(**res.variation)
            except ValidationError:
                variation = None
            else:
                variation = res.variation
        elif variation_type == "RelativeCopyNumber":
            try:
                RelativeCopyNumber(**res.variation)
            except ValidationError:
                variation = None
            else:
                variation = res.variation
        else:
            raise Exception(f"{variation_type} not supported in "
                            f"Variation Normalization")
        return variation



"""Module for creating VRS objects"""
from typing import List, Optional, Tuple, Union, Dict

from ga4gh.vrs import models, normalize
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.core import ga4gh_identify
from bioutils.accessions import coerce_namespace
from uta_tools.data_sources import SeqRepoAccess


class VRS:
    """VRS class"""

    def __init__(self, dp: SeqRepoDataProxy,
                 seqrepo_access: SeqRepoAccess) -> None:
        """Initialize VRS class

        :param SeqRepoAccess seqrepo_access: Access to SeqRepo
        :param SeqRepoDataProxy dp: SeqRepoDataProxy
        """
        self.seqrepo_access = seqrepo_access
        self.dp = dp

    @staticmethod
    def get_ival_start_end(
            coordinate: str, start: int, end: int, cds_start: int,
            errors: List) -> Optional[Tuple[int, int]]:
        """Get ival_start and ival_end coordinates.

        :param str coordinate: Coordinate used. Must be either `p`, `c`, or `g`
        :param int start: Start position change
        :param int end: End position change
        :param int cds_start: Coding start site
        :param List errors: List of errors
        :return: Tuple[ival_start, ival_end]
        """
        try:
            start = int(start)
            if end is None:
                end = start
            end = int(end)
        except (ValueError, TypeError):
            errors.append("Start/End must be valid ints")
            return None

        if coordinate == "c":
            if cds_start:
                start += cds_start
                end += cds_start
        return start, end

    @staticmethod
    def get_start_indef_range(start: int) -> models.IndefiniteRange:
        """Return indefinite range given start coordinate

        :param int start: Start position (assumes 1-based)
        :return: Indefinite range model
        """
        return models.IndefiniteRange(value=start - 1, comparator="<=",
                                      type="IndefiniteRange")

    @staticmethod
    def get_end_indef_range(end: int) -> models.IndefiniteRange:
        """Return indefinite range given end coordinate

        :param int end: End position (assumes 1-based)
        :return: Indefinite range model
        """
        return models.IndefiniteRange(value=end, comparator=">=",
                                      type="IndefiniteRange")

    @staticmethod
    def get_ival_certain_range(start1: int, start2: int, end1: int,
                               end2: int) -> models.SequenceInterval:
        """Return sequence interval

        :param int start1: Start left pos (assumes 1-based)
        :param int start2: Start right pos (assumes 1-based)
        :param int end1: End left pos (assumes 1-based)
        :param int end2: End right pos (assumes 1-based)
        :return: Sequence Interval model
        """
        return models.SequenceInterval(
            start=models.DefiniteRange(min=start1 - 1, max=start2 - 1,
                                       type="DefiniteRange"),
            end=models.DefiniteRange(min=end1 + 1, max=end2 + 1,
                                     type="DefiniteRange"),
            type="SequenceInterval"
        )

    @staticmethod
    def get_sequence_loc(
            ac: str, interval: models.SequenceInterval) -> models.Location:
        """Return VRS location

        :param str ac: Accession
        :param models.SequenceInterval interval: VRS sequence interval
        :return: VRS Location model
        """
        return models.SequenceLocation(
            sequence_id=coerce_namespace(ac),
            interval=interval, type="SequenceLocation")

    def vrs_allele(self, ac: str, interval: models.SequenceInterval,
                   sstate: Union[models.LiteralSequenceExpression,
                                 models.DerivedSequenceExpression,
                                 models.RepeatedSequenceExpression],
                   alt_type: str, errors: List) -> Optional[Dict]:
        """Create a VRS Allele object.

        :param str ac: Accession
        :param SequenceInterval interval: Sequence Interval
        :param sstate: State
        :type sstate: models.LiteralSequenceExpression or
            models.DerivedSequenceExpression or
            models.RepeatedSequenceExpression
        :param str alt_type: Type of alteration
        :param List errors: List of errors
        :return: VRS Allele object represented as a Dict
        """
        try:
            location = self.get_sequence_loc(ac, interval)
        except ValueError as e:
            errors.append(f"Unable to get sequence location: {e}")
            return None
        allele = models.Allele(location=location, state=sstate, type="Allele")
        # Ambiguous regions do not get normalized
        if alt_type not in ["uncertain_deletion", "uncertain_duplication",
                            "duplication_range", "deletion_range"]:
            try:
                allele = normalize(allele, self.dp)
            except (KeyError, AttributeError) as e:
                errors.append(f"vrs-python unable to normalize allele: {e}")
                return None

        if not allele:
            errors.append("Unable to get allele")
            return None

        seq_id = self.dp.translate_sequence_identifier(
            allele.location.sequence_id._value, "ga4gh")[0]
        allele.location.sequence_id = seq_id
        allele.location._id = ga4gh_identify(allele.location)
        allele._id = ga4gh_identify(allele)
        return allele.as_dict()

    def to_vrs_allele(
            self, ac: str, start: int, end: int, coordinate: str,
            alt_type: str, errors: List, cds_start: int = None,
            alt: str = None) -> Optional[Dict]:
        """Translate accession and position to VRS Allele Object.

        :param str ac: Accession
        :param int start: Start position change
        :param int end: End position change
        :param str coordinate: Coordinate used. Must be either `p`, `c`, or `g`
        :param str alt_type: Type of alteration
        :param List errors: List of errors
        :param int cds_start: Coding start site
        :param str alt: Alteration
        :return: VRS Allele Object
        """
        ival_coords = self.get_ival_start_end(coordinate, start, end,
                                              cds_start, errors)
        if not ival_coords:
            return None
        if ival_coords[0] > ival_coords[1]:
            ival_end, ival_start = ival_coords
        else:
            ival_start, ival_end = ival_coords

        # Right now, this follows HGVS conventions
        # This will change once we support other representations
        if alt_type == "insertion":
            state = alt
            ival_end = ival_start
        elif alt_type in ["substitution", "deletion", "delins",
                          "silent_mutation", "nonsense"]:
            if alt_type == "silent_mutation":
                state, _ = self.seqrepo_access.get_reference_sequence(ac, ival_start)
                if state is None:
                    errors.append(f"Unable to get sequence on {ac} from "
                                  f"{ival_start}")
                    return None
            else:
                state = alt or ""
            ival_start -= 1
        elif alt_type == "duplication":
            ref, _ = self.seqrepo_access.get_reference_sequence(
                ac, ival_start, ival_end + 1)
            if ref is not None:
                state = ref + ref
            else:
                errors.append(f"Unable to get sequence on {ac} from "
                              f"{ival_start} to {ival_end + 1}")
                return None
            ival_start -= 1
        else:
            errors.append(f"alt_type not supported: {alt_type}")
            return None

        interval = models.SequenceInterval(
            start=models.Number(value=ival_start, type="Number"),
            end=models.Number(value=ival_end, type="Number"),
            type="SequenceInterval")
        sstate = models.LiteralSequenceExpression(sequence=state,
                                                  type="LiteralSequenceExpression")
        return self.vrs_allele(ac, interval, sstate, alt_type, errors)

    def to_vrs_allele_ranges(
            self, ac: str, coordinate: str, alt_type: str, errors: List,
            ival: models.SequenceInterval) -> Optional[Dict]:
        """Translate variation ranges to VRS Allele Object.

        :param str ac: Accession
        :param str coordinate: Coordinate used. Must be either `p`, `c`, or `g`
        :param str alt_type: Type of alteration
        :param List errors: List of errors
        :param models.SequenceInterval ival: Sequence Interval
        :return: VRS Allele object
        """
        if coordinate == "c":
            # TODO: Once we add support for ranges on c. coord
            return None
        if alt_type in ["uncertain_deletion", "uncertain_duplication",
                        "duplication_range", "deletion_range"]:
            sstate = models.LiteralSequenceExpression(
                sequence="", type="LiteralSequenceExpression"
            )
        else:
            errors.append("No state")
            return None

        return self.vrs_allele(ac, ival, sstate, alt_type, errors)



"""A module for Coding DNA Silent Mutation Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import CodingDNASilentMutationToken
from .single_nucleotide_variation_base import SingleNucleotideVariationBase


class CodingDNASilentMutation(SingleNucleotideVariationBase):
    """Class for tokenizing Silent Mutation at the coding dna
    reference sequence.
    """

    def return_token(self, params: Dict) -> Optional[CodingDNASilentMutationToken]:
        """Return coding DNA silent mutation token."""
        if self.sub["coordinate_type"] == "c" and \
                self.sub["ref_nucleotide"] is None and \
                self.sub["new_nucleotide"] == "=":
            return CodingDNASilentMutationToken(**params)



"""Module for testing Coding DNA Substitution Validator."""
import unittest

from variation.validators import CodingDNASubstitution
from variation.classifiers import CodingDNASubstitutionClassifier
from .validator_base import ValidatorBase


class TestCodingDNASubstitutionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Coding DNA Substitution Validator."""

    def validator_instance(self):
        """Return coding DNA substitution instance."""
        return CodingDNASubstitution(*self.params)

    def classifier_instance(self):
        """Return the coding DNA substitution classifier instance."""
        return CodingDNASubstitutionClassifier()

    def fixture_name(self):
        """Return the fixture name for coding DNA substitution."""
        return "coding_dna_substitution"



"""The module for Polypeptide Truncation Validation."""
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import Token
from .polypeptide_sequence_variation_base import PolypeptideSequenceVariationBase


class PolypeptideTruncation(PolypeptideSequenceVariationBase):
    """The Polypeptide Truncation Validator class."""

    def variation_name(self) -> str:
        """Return the variation name."""
        return "polypeptide truncation"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Polypeptide Truncation"""
        return t.token_type == "PolypeptideTruncation"

    def validates_classification_type(
        self, classification_type: ClassificationType
    ) -> bool:
        """Return whether or not the classification type is polypeptide
        truncation.
        """
        return classification_type == ClassificationType.POLYPEPTIDE_TRUNCATION



"""Module for schemas used throughout the app"""
from enum import Enum


class Endpoint(str, Enum):
    """Define contrains for endpoint names"""

    TO_VRS = "toVRS"
    NORMALIZE = "normalize"
    TRANSLATE_IDENTIFIER = "translate_identifier"
    GNOMAD_VCF_TO_PROTEIN = "gnomad_vcf_to_protein"
    CANONICAL_SPDI_TO_CAT_VAR = "canonical_spdi_to_categorical_variation"
    TRANSLATE_FROM = "translate_from"
    HGVS_TO_ABSOLUTE_CN = "hgvs_to_absolute_copy_number"
    HGVS_TO_RELATIVE_CN = "hgvs_to_relative_copy_number"



"""Module for Tokenization."""
from abc import ABC, abstractmethod
from typing import Optional

from variation.schemas.token_response_schema import Token


class Tokenizer(ABC):
    """The tokenizer class."""

    @abstractmethod
    def match(self, input_string: str) -> Optional[Token]:
        """Return tokens that match the input string."""
        raise NotImplementedError



"""Module for vrs-python translator endpoint response schema"""
from enum import Enum
from typing import Optional, List, Union

from ga4gh.vrsatile.pydantic.vrs_models import Allele
from pydantic import BaseModel
from pydantic.types import StrictStr

from variation.schemas.normalize_response_schema import ServiceMeta


class VrsPythonMeta(BaseModel):
    """Metadata regarding vrs-python dependency"""

    name = "vrs-python"
    version: str
    url = "https://github.com/ga4gh/vrs-python"


class TranslateFromFormat(str, Enum):
    """Enums for formats that vrs-python can translate from"""

    HGVS = "hgvs"
    BEACON = "beacon"
    GNOMAD = "gnomad"
    SPDI = "spdi"


class TranslateToFormat(str, Enum):
    """Enums for formats that vrs-python can translate to"""

    HGVS = "hgvs"
    SPDI = "spdi"


class TranslateToQuery(BaseModel):
    """Query fields for Translate To Service"""

    variation: Allele
    fmt: TranslateToFormat


class TranslateFromQuery(BaseModel):
    """Query fields for Translate From Service"""

    variation: StrictStr
    fmt: Optional[TranslateFromFormat] = None


class TranslateService(BaseModel):
    """Response schema for vrs-python translator endpoints"""

    query: Union[TranslateFromQuery, TranslateToQuery]
    warnings: Optional[List[StrictStr]]
    service_meta_: ServiceMeta
    vrs_python_meta_: VrsPythonMeta


class TranslateFromService(TranslateService):
    """Response schema for vrs-python translate from endpoint"""

    variation: Optional[Allele] = None


class TranslateToService(TranslateService):
    """Response schema for vrs-python translate to endpoint"""

    variations: List[StrictStr]


class ErrorResponse(BaseModel):
    """Response for Translate services when request body is invalid"""

    errors: List[StrictStr]



"""A module for testing Protein Deletion Tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import ProteinDeletion
from .tokenizer_base import TokenizerBase


class TestProteinDeletionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Protein Deletion Tokenization."""

    def tokenizer_instance(self):
        """Return Protein Deletion instance."""
        return ProteinDeletion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return protein deletion token type."""
        return "ProteinDeletion"

    def fixture_name(self):
        """Return the fixture name for protein deletion."""
        return "protein_deletion"



"""A module for Coding DNA Deletion Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import CodingDNADeletionToken
from variation.tokenizers.deletion_base import DeletionBase


class CodingDNADeletion(DeletionBase):
    """Class for tokenizing Deletion at the coding dna reference sequence."""

    def return_token(self, params: Dict) -> Optional[CodingDNADeletionToken]:
        """Return coding DNA Deletion token."""
        if self.parts["coordinate_type"] == "c":
            return CodingDNADeletionToken(**params)



"""Module for testing Protein DelIns Classifier."""
import unittest

from variation.classifiers import ProteinDelInsClassifier
from .classifier_base import ClassifierBase


class TestProteinDelInsClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Protein DelIns Classifier."""

    def classifier_instance(self):
        """Return ProteinDelInsClassifier instance."""
        return ProteinDelInsClassifier()

    def fixture_name(self):
        """Return ProteinDelInsClassifier fixture name."""
        return "protein_delins"
